from typing import Tuple

import numpy as np
import numba as nb
from numba import njit, prange

# Numba-optimised functions
@njit
def ols_estim(y: np.array, X: np.array) -> Tuple[np.array, float]:
    """Estimates the beta and R-squared of an OLS regression

    Assumes that the intercept is included in X.

    Args:
        y (np.array): Dependent variable observations (dimension T x 1)
        X (np.array): Explanatory variables observations (dimension T x K)

    Returns:
        Tuple[np.array, float]: Beta coefficients (dimension K x 1), R-squared
    """
    p = X.shape[1]
    try:
        XtXinv = np.linalg.inv(X.T @ X)

        # Estimate coefficients
        beta = XtXinv @ (X.T @ y)

        # Compute residuals
        e = y - X @ beta
        # Compute R^2 and adj R^2
        r2 = 1 - np.sum(e**2) / np.sum((y - np.mean(y)) ** 2)

    except Exception:
        beta = np.full((p, 1), np.nan)
        r2 = np.nan

    return beta, r2


@njit
def ols_estim_se(y: np.array, X: np.array) -> Tuple[np.array, np.array, float]:
    """Estimates the beta, White standard errors, and R-squared of an OLS regression

    Assumes that the intercept is included in X.

    Args:
        y (np.array): Dependent variable observations (dimension T x 1)
        X (np.array): Explanatory variables observations (dimension T x K)

    Returns:
        Tuple[np.array, np.array, float]: Beta coefficients (dimension K x 1), standard errors (dimension K x K), R-squared
    """
    p = X.shape[1]
    try:
        XtXinv = np.linalg.inv(X.T @ X)

        # Estimate coefficients
        beta = XtXinv @ (X.T @ y)

        # Compute residuals
        e = y - X @ beta
        # Compute R^2 and adj R^2
        r2 = 1 - np.sum(e**2) / np.sum((y - np.mean(y)) ** 2)

        # Compute standard errors (White)
        Sigma_HC = XtXinv @ (X.T @ np.diag(e.flatten() ** 2) @ X) @ XtXinv
        se = np.diag(Sigma_HC) ** 0.5

    except Exception:
        beta = np.full((p, 1), np.nan)
        se = np.full((p,), np.nan)
        r2 = np.nan

    return beta, se, r2


@njit
def compute_unbiasedness_reg_jit(
    data: nb.typed.typedlist.List,
) -> Tuple[np.array, np.array]:
    """Estimates the betas and R-squared of an unbiasedness regression

    Assumes that the intercept is included in X, and that there is only one explanatory variable.
    The beta coefficient is the coefficient on the explanatory variable.

    Args:
        data (nb.typed.typedlist.List): A list of tuples (y, X) where y is the dependent variable observations (dimension T x 1) and X is the explanatory variables observations (dimension T x K)

    Returns:
        Tuple[np.array, np.array, np.array]: Beta coefficients (dimension N x 1), R-squared (dimension N x 1)
    """
    N = len(data)
    beta = np.empty(N)
    r2 = np.empty(N)
    for i in prange(N):
        estim_i = ols_estim(data[i][0], data[i][1])
        beta[i] = estim_i[0][1][0]
        r2[i] = estim_i[1]
    return beta, r2


@njit
def compute_unbiasedness_reg_se_jit(
    data: nb.typed.typedlist.List,
) -> Tuple[np.array, np.array, np.array]:
    """Estimates the betas, White standard errors, and R-squared of an unbiasedness regression

    Assumes that the intercept is included in X, and that there is only one explanatory variable.
    The beta coefficient (and associated standard errors) is the coefficient on the explanatory variable.

    Args:
        data (nb.typed.typedlist.List): A list of tuples (y, X) where y is the dependent variable observations (dimension T x 1) and X is the explanatory variables observations (dimension T x K)

    Returns:
        Tuple[np.array, np.array, np.array]: Beta coefficients (dimension N x 1), standard errors (dimension N x 1), R-squared (dimension N x 1)
    """
    N = len(data)
    beta = np.empty(N)
    se = np.empty(N)
    r2 = np.empty(N)
    for i in prange(N):
        estim_i = ols_estim_se(data[i][0], data[i][1])
        beta[i] = estim_i[0][1][0]
        se[i] = estim_i[1][1]
        r2[i] = estim_i[2]
    return beta, se, r2



from typing import Iterable, Union, Callable, Tuple, List, Any, Dict
from dataclasses import dataclass
from enum import Enum

import numpy as np
import pandas as pd


class TableFileType(Enum):
    LATEX = 1
    MARKDOWN = 2


@dataclass
class TableFormat:
    file_type: TableFileType
    sep: str
    newline: str
    line_start: str
    line_end: str
    empty_col: str
    empty_col_header: str
    footer: str
    col_name_formatter: Callable[Tuple[str, int], str]
    coef_formatter: Callable[[Union[float, np.float64]], str]
    pval_formatter: Callable[[Union[float, np.float64]], str]
    header: Callable[[Union[int, int]], str]
    header_sep: Callable[[Union[int, int]], str]


TEX_EMPTY_COL = " " * 3
MD_EMPTY_COL = " " * 10


def tex_header_sep(nindex: int, col_names_list: List[List[Any]]) -> str:
    i = nindex
    out = ""
    for c in col_names_list:
        out += f"\\cmidrule(lr){'{'}{i+1}-{i+c[1]}{'}'}"
        i += c[1]
    return out + "\n"


latex = TableFormat(
    file_type=TableFileType.LATEX,
    sep=" & ",
    newline="\\\\\n",
    line_start="",
    line_end="",
    empty_col=TEX_EMPTY_COL,
    empty_col_header="{}",
    footer="\\hline\n\\hline\n\\end{tabular}\n",
    col_name_formatter=lambda c, n: "\\multicolumn{" + f"{n}" + "}{c}{" + c + "}"
    if n > 1
    else c,
    coef_formatter=lambda x: TEX_EMPTY_COL if np.isnan(x) else f"{x:0.3f}",
    pval_formatter=lambda x: TEX_EMPTY_COL
    if np.isnan(x)
    else f"\\emph{'{'}{x:0.3f}{'}'}",
    header=lambda nindex, ncols: (
        "\\begin{tabular}{" + "r" * nindex + "c" * ncols + "}\n\\hline\n\\hline\n"
    ),
    header_sep=tex_header_sep,
)

markdown = TableFormat(
    file_type=TableFileType.MARKDOWN,
    sep=" | ",
    newline="\n",
    line_start="| ",
    line_end=" |",
    empty_col=MD_EMPTY_COL,
    empty_col_header=" ",
    footer="\n",
    col_name_formatter=lambda c, n: c + " | " * (n - 1),
    coef_formatter=lambda x: MD_EMPTY_COL if np.isnan(x) else f"{x:10.3f}",
    pval_formatter=lambda x: MD_EMPTY_COL if np.isnan(x) else f"_{x:0.3f}_",
    header=lambda nindex, ncols: "",
    header_sep=lambda nindex, col_names_list: ("|:---" * nindex)
    + "|"
    + "---:|" * sum(x[1] for x in col_names_list)
    + "\n",
)


# Utility function to format windows labels
def win_lbl(w: Tuple[int, int]) -> str:
    return f"$[{w[0]},{w[1]}]$"


# Utility function for maps to keep ordering constant with stack/unstack
def map_ordering(l: List[str]) -> Tuple[Dict[str, int], Dict[int, str]]:
    unique = []
    for x in l:
        if x not in unique:
            unique.append(x)
    return {x: i for i, x in enumerate(unique)}, dict(enumerate(unique))


class UnbiasednessTable:
    def __init__(self, df: pd.DataFrame):
        self.df = df

    @staticmethod
    def _format_index(index: Union[str, Tuple[str]], sep: str) -> Tuple[str, bool]:
        if isinstance(index, str):
            index = (index,)
        return sep.join(index), index[-1] != ""

    @staticmethod
    def _col_names_row(nindex, col_names_list, format: TableFormat):
        cols_headers = [format.empty_col_header] * nindex
        for c, n in col_names_list:
            cols_headers.append(format.col_name_formatter(c, n))
        return format.sep.join(cols_headers) + format.line_end

    @staticmethod
    def _df_to_col_names(cols: Iterable[str]) -> List[List[List[Any]]]:
        out = []
        for i in range(cols.nlevels):
            vals = cols.get_level_values(i)
            col_names_list = []
            for j, v in enumerate(vals):
                if not col_names_list or col_names_list[-1][0] != v:
                    col_names_list.append([v, 1])
                elif (
                    i > 0
                    and cols.get_level_values(0)[j] != cols.get_level_values(0)[j - 1]
                ):
                    col_names_list.append([v, 1])
                else:
                    col_names_list[-1][1] += 1
            out.append(col_names_list)
        return out

    @staticmethod
    def _copy_df_and_rename_cols(df, cols, windows, label):
        df = df[[c for c in cols if c in df.columns]].copy()
        df.columns = pd.MultiIndex.from_tuples(
            zip([label] * len(df.columns), df.columns)
        )
        return df.loc[windows]

    @staticmethod
    def _format_df(df: pd.DataFrame, long: bool = False) -> pd.DataFrame:
        df = df.reset_index().copy()

        sample = "Sample" in df.columns.names

        if long and sample:
            raise NotImplementedError(
                "Long format not implemented for multiple samples"
            )
        elif long:
            # Need to rename periods and windows to keep ordering
            # (stack/unstack will change it)
            win_map, win_map_inv = map_ordering(df.Window.unique())
            df["Window"] = df.Window.map(win_map)
            df = df.set_index(["Window", "Value"]).T.reset_index()
            per_map, per_map_inv = map_ordering(df.Period.unique())
            df["Period"] = df.Period.map(per_map)
            df = df.set_index(["Period", "Metric"]).T

            df = df.unstack(level="Value").stack(level="Metric").T.reset_index()
            df["Period"] = df.Period.map(per_map_inv)
            df = df.set_index(["Period", "Value"])

            df = df.T.reset_index()
            df["Window"] = df.Window.map(win_map_inv)
            df = df.set_index(["Window", "Metric"]).T.reset_index(level=1)

            df = df.reset_index()
            df.loc[df.Value != "1_coef", "Period"] = ""
            df = df.set_index("Period")
        else:
            df.loc[df.Value != "1_coef", "Window"] = ""
            df = df.set_index("Window")
        del df["Value"]
        df.index.names = [""]

        return df

    @staticmethod
    def _build_table(
        df,
        format: TableFormat = latex,
        header=True,
        col_header=True,
        footer=True,
        index=True,
        long=False,
    ):
        # Format df properly and get column names.
        df = UnbiasednessTable._format_df(df, long=long)
        ncols = df.shape[1]
        col_names_list = UnbiasednessTable._df_to_col_names(df.columns)

        nindex = 1 if index else 0

        out = ""
        if header:
            out += format.header(nindex=nindex, ncols=ncols)

        if col_header:
            out += (
                format.line_start
                + UnbiasednessTable._col_names_row(nindex, col_names_list[0], format)
                + format.newline
            )
            out += format.header_sep(nindex, col_names_list[0])
            if len(col_names_list) > 1:
                for c in col_names_list[1:]:
                    out += (
                        format.line_start
                        + UnbiasednessTable._col_names_row(nindex, c, format)
                        + format.newline
                    )

        for r in df.iterrows():
            out += format.line_start
            index_val, is_coef = UnbiasednessTable._format_index(
                index=r[0], sep=format.sep
            )
            if index:
                out += index_val + format.sep

            if is_coef:
                out += format.sep.join([format.coef_formatter(x) for x in r[1]])
            else:
                out += format.sep.join([format.pval_formatter(x) for x in r[1]])

            out += format.line_end + format.newline
        if footer:
            out += format.footer
        elif format.file_type == TableFileType.LATEX:
            # Remove last line ending
            new_end = len(format.line_end + format.newline)
            out = out[:-new_end]
        return out

    def to_latex(
        self,
        fn,
        header: bool = True,
        col_header: bool = True,
        footer: bool = True,
        index: bool = True,
        long: bool = False,
    ):
        with open(fn, "w") as file:
            file.write(
                self._build_table(
                    self.df,
                    format=latex,
                    header=header,
                    col_header=col_header,
                    footer=footer,
                    index=index,
                    long=long,
                )
            )

    def to_markdown(
        self,
        fn,
        header: bool = True,
        col_header: bool = True,
        footer: bool = True,
        index: bool = True,
        long: bool = False,
    ):
        with open(fn, "w") as file:
            file.write(
                self._build_table(
                    self.df,
                    format=markdown,
                    header=header,
                    col_header=col_header,
                    footer=footer,
                    index=index,
                    long=long,
                )
            )

    def _repr_markdown_(self):
        return self._build_table(
            self.df, format=markdown, header=True, footer=True, long=True
        )

    @staticmethod
    def _compile_unbiasedness_regs_df(
        res,
        period_labels,
        windows,
        bs_vals: Iterable[pd.DataFrame] = (),
        add_h0=True,
        excess_only=False,
    ):
        base_cols = ["_Delta_R2", "_Delta_Excess_R2"]
        if excess_only:
            base_cols = ["_Delta_Excess_R2"]
        cols = [x[0] + s for x in period_labels for s in base_cols]

        table_out = (
            pd.concat(
                [
                    UnbiasednessTable._copy_df_and_rename_cols(
                        res, cols, windows, "0_Estim"
                    )
                ]
                + [
                    UnbiasednessTable._copy_df_and_rename_cols(
                        b, cols, windows, f"{i+1}_BS{i+1}"
                    )
                    for i, b in enumerate(bs_vals)
                ],
                axis=1,
            )
            .stack(level=0)[cols]
            .reset_index(level=2, drop=True)
        )

        val_labels = ["1_coef"] + [f"bs{i+1}" for i in range(len(bs_vals))]

        table_out.index = pd.MultiIndex.from_arrays(
            [
                [win_lbl(w) for w in windows for _ in range(len(val_labels))],
                val_labels * len(windows),
            ],
            names=["Window", "Value"],
        )

        table_out.columns = (
            pd.MultiIndex.from_arrays(
                [
                    [p[1] for p in period_labels],
                    ["Excess $\Delta R^2$"] * len(period_labels),
                ],
                names=["Period", "Metric"],
            )
            if excess_only
            else pd.MultiIndex.from_arrays(
                [
                    [p[1] for p in period_labels for _ in (0, 1)],
                    list(["$\Delta R^2$", "Excess $\Delta R^2$"] * len(period_labels)),
                ],
                names=["Period", "Metric"],
            )
        )

        if add_h0:
            table_out.insert(
                0,
                "$H_0$",
                pd.Series({win_lbl(w): 1 / (w[1] - w[0] + 1) for w in windows}),
            )
        return table_out

    @staticmethod
    def from_unbiasedness_regs(
        res,
        period_labels,
        windows,
        bs_vals: Iterable[pd.DataFrame] = (),
        add_h0=True,
        excess_only=False,
    ):
        table_out = UnbiasednessTable._compile_unbiasedness_regs_df(
            res,
            period_labels,
            windows,
            bs_vals=bs_vals,
            add_h0=add_h0,
            excess_only=excess_only,
        )

        return UnbiasednessTable(table_out)

    @staticmethod
    def _compile_subsample_df(subsample_regs, period_labels, windows):
        cols = [x[0] + s for x in period_labels for s in ["_Delta_Excess_R2"]]

        tables_samples = []
        for x in subsample_regs:
            res = subsample_regs[x][0]
            bs_vals = subsample_regs[x][1:]

            val_labels = ["1_coef"] + [f"bs{i+1}" for i in range(len(bs_vals))]
            table_sample = (
                pd.concat(
                    [
                        UnbiasednessTable._copy_df_and_rename_cols(
                            res, cols, windows, "0_Estim"
                        )
                    ]
                    + [
                        UnbiasednessTable._copy_df_and_rename_cols(
                            bs, cols, windows, f"{i+1}_BS{i+1}"
                        )
                        for i, bs in enumerate(bs_vals)
                    ],
                    axis=1,
                )
                .stack(level=0)[cols]
                .reset_index(level=2, drop=True)
            )
            table_sample.index = pd.MultiIndex.from_arrays(
                [
                    [win_lbl(w) for w in windows for _ in range(len(val_labels))],
                    val_labels * len(windows),
                ],
                names=["Window", "Value"],
            )
            tables_samples.append(table_sample)

        table_out = pd.concat(tables_samples, axis=1)

        table_out.columns = pd.MultiIndex.from_arrays(
            [
                [s for s in subsample_regs for _ in range(len(period_labels))],
                ["Excess $\Delta R^2$"] * len(period_labels) * len(subsample_regs),
                [p[1] for p in period_labels] * len(subsample_regs),
            ]
        )
        table_out.columns.names = ["Sample", "Metric", "Period"]

        return table_out

    @staticmethod
    def from_subsample_regs(subsample_regs, period_labels, windows):
        df = UnbiasednessTable._compile_subsample_df(
            subsample_regs, period_labels, windows
        )

        return UnbiasednessTable(df=df)



from typing import Tuple, Iterable, Dict

import pandas as pd
import numba as nb

from .numba_jit import compute_unbiasedness_reg_se_jit, compute_unbiasedness_reg_jit

from .windows import get_return_windows, combine_window


def unbiasedness_reg(
    df: pd.DataFrame, use_se: bool = True, include_vr: bool = False
) -> pd.DataFrame:
    """Estimates the betas and R-squared of an unbiasedness regression

    Args:
        df (_type_): DatFrame with columns "t", "CRet", "CRet_End"
        use_se (bool, optional): Include standard errors in the output. Defaults to True.
        include_vr (bool, optional): Include variance ratio in the output. Defaults to False.

    Returns:
        pd.DataFrame: beta coefficients and R-squared associated with each time period
    """
    df = df.copy()
    df["cst"] = 1

    windows = {
        l: (x[["CRet_End"]].values, x[["cst", "CRet"]].values)
        for l, x in df.groupby("t")
    }
    keys = list(windows.keys())

    if not keys:
        return pd.DataFrame(
            columns=["beta", "se", "r2"] if use_se else ["beta", "r2"]
        ).T

    data = nb.typed.List([windows[k] for k in keys])

    if use_se:
        out = pd.DataFrame(
            compute_unbiasedness_reg_se_jit(data),
            columns=keys,
            index=["beta", "se", "r2"],
        ).T
    else:
        out = pd.DataFrame(
            compute_unbiasedness_reg_jit(data), columns=keys, index=["beta", "r2"]
        ).T

    out.index.name = "t"

    if include_vr:
        out["vr"] = df.groupby("t")["CRet"].var() / df.groupby("t")["CRet_End"].var()

    return out


def estim_unbiasedness_windows(
    win_dfs: Dict[str, pd.DataFrame],
    periods: Iterable[Tuple[int, int]],
    use_se: bool = True,
    include_vr: bool = False,
) -> pd.DataFrame:
    regs = {
        w: unbiasedness_reg(win_df, use_se=use_se, include_vr=include_vr)
        for w, win_df in win_dfs.items()
    }
    return pd.DataFrame({w: combine_window(w, regs[w], periods) for w in win_dfs}).T


def estim_unbiasedness_regs(
    returns: pd.Series,
    dates: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    periods: Iterable[Tuple[int, int]],
    use_se: bool = False,
    include_vr: bool = False,
    shuffle: bool = False,
) -> pd.DataFrame:
    # Keep only dates that are in the returns
    dates = returns.index.intersection(dates)

    win_dfs = get_return_windows(returns, dates, windows, shuffle=shuffle)
    return estim_unbiasedness_windows(
        win_dfs, periods, use_se=use_se, include_vr=include_vr
    )



from typing import Tuple, Iterable, Dict, Union
from collections import ChainMap

import pandas as pd
import numpy as np


def resample_windows(
    win_df: Dict[Tuple[int, int] | str, pd.DataFrame]
) -> Dict[Tuple[int, int] | str, pd.DataFrame]:
    """Resample observations with replacement.

    Args:
        win_df (Dict[Tuple[int, int], pd.DataFrame]): Dictionnary of dataframes with observations
        for each window.

    Returns:
        pd.DataFrame: Resampled observations.

    We assume that each dataframe has the same event dates. Each dataframe contains
    observations for a given window for multiple events. We resample with replacement
    by event date, i.e. we pick random event dates, and then pick all observations
    for those event dates. To avoid issues with duplicate event dates, we replace
    the resampled event dates with a simple index.

    The event dates are assumed to be in the column `index`.
    """
    if not win_df:
        raise ValueError("win_df is empty")

    try:
        first_key = next(iter(win_df))
        if type(win_df[first_key]) is not pd.DataFrame:
            raise ValueError("win_df must be a dictionary of dataframes")
        dates = win_df[first_key]["index"].unique()
    except AttributeError as e:
        raise ValueError("win_df must be a dictionary of dataframes") from e

    # Sample index with replacement
    dates = np.random.choice(dates, size=len(dates), replace=True)

    # Resample each dataframe
    return {k: v.set_index("index").loc[dates].reset_index() for k, v in win_df.items()}


def shuffle_window(
    df: pd.DataFrame, group_col: str = "index", shuffle_col: str = "ret"
) -> pd.Series:
    """Shuffle a column within groups.

    Args:
        df (pd.DataFrame): Dataframe with a column to shuffle.
        group_col (str, optional): Column to group by. Defaults to "index".
        shuffle_col (str, optional): Column to shuffle. Defaults to "ret".

    Returns:
        pd.Series: Shuffled column
    """
    idx = df.index
    s_out = df.groupby(group_col)[shuffle_col].sample(frac=1, replace=False)
    s_out.index = idx
    return s_out


def combine_return_windows(
    df: pd.DataFrame, windows: Iterable[Tuple[int, int]]
) -> Dict[Tuple[int, int], pd.DataFrame]:
    """Combines the returns in the windows into a single cumulative returns series

    Args:
        df (pd.DataFrame): DataFrame with columns "t", "ret", "index"
        windows (Iterable[Tuple[int, int]]): List of tuples with the start and end of each window

    Returns:
        Dict[Tuple[int, int], pd.DataFrame]: Augmented DataFrame for each window


    Output DataFrames have original columns plus "CRet" and "CRet_End".
    """
    df = df.sort_values(["index", "t"])
    out = {w: df[(df["t"] >= w[0]) & (df["t"] <= w[1])].copy() for w in windows}
    for w in out:
        out[w]["CRet"] = out[w].groupby("index")["ret"].cumsum()
        out[w]["CRet_End"] = out[w].groupby("index")["CRet"].transform("last")
        out[w] = out[w].reset_index(drop=True)

    return out


def merge_events_returns(
    returns: pd.Series, events: pd.DatetimeIndex, windows: Iterable[Tuple[int, int]]
) -> pd.DataFrame:
    """Merges the returns and events into a single DataFrame

    Args:
        returns (pd.Series): Series of returns
        events (pd.DatetimeIndex): DatetimeIndex of events

    Returns:
        pd.DataFrame: DataFrame with columns "t", "ret", "date_evt", and "date_t"
    """
    min_start = min(window[0] for window in windows)
    max_end = max(window[1] for window in windows)

    # Create longest possible window first
    returns_df = pd.DataFrame(returns)
    returns_df.columns = ["ret"]
    returns_df.index.name = "date"
    returns_df["date_t"] = pd.Series(range(len(returns)), index=returns.index)

    evt_dates_df = returns_df.loc[returns_df.index.intersection(events), ["date_t"]]
    evt_dates_df.index.name = "date"

    window_ts = pd.DataFrame({"t": range(min_start, max_end + 1)})

    evt_win_df = (
        evt_dates_df.assign(key=1)
        .reset_index()
        .merge(window_ts.assign(key=1), on="key")
        .drop(columns="key")
        .set_index("date")
    )
    evt_win_df["date_t"] = evt_win_df["date_t"] + evt_win_df["t"]

    win_df = pd.merge(
        evt_win_df.reset_index(),
        returns_df.reset_index()[["date", "date_t", "ret"]],
        on="date_t",
        suffixes=("_evt", "_obs"),
    )
    del win_df["date_t"]
    return win_df


def get_return_windows(
    returns: pd.Series,
    events: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    shuffle: bool = False,
) -> Dict[Tuple[int, int], pd.DataFrame]:
    """
    Computes return windows around a set of events.

    Args:
        returns: A pandas series of returns.
        events: A pandas datetime index of event dates.
        windows: An iterable of tuples representing the start and end offsets (in days)
        for each window.
            The start offset can be negative to indicate a window that starts before the
            event date.
            The end offset can be 'T1' to indicate a window that ends at the next event
            date.
        shuffle: Whether to shuffle the returns within each window for bootstrap, by
        default False.

    Returns:
        A dictionary of dataframes, where each key is a tuple representing the start and
        end offsets
        for a window, and each value is a dataframe containing the returns within that
        window for each event.
        The dataframes have the following columns:
            - t: The offset from the event date
            - ret: The return at that offset
            - index: The event date or resample id if resampling
            - date_obs: The date of the return
            - "CRet": The cumulative return at that offset
            - "CRet_End": The cumulative return at the end of the window
    """
    win_df = merge_events_returns(returns, events, windows)

    # Shuffle window for bootstrap
    if shuffle:
        win_df["ret"] = shuffle_window(win_df, group_col="index", shuffle_col="ret")

    win_df = win_df.rename(columns={"date_evt": "index"})

    return combine_return_windows(win_df, windows)


def period_prefix(p: Tuple[Union[int, str], int]) -> str:
    """Generate a prefix for a period.

    Args:
        p (Tuple[Union[int, str], int]): Period start and end.

    Returns:
        str: Prefix for the period.
    """
    p_label = f"p{p[0]}_p{p[1]}" if p[0] != p[1] else f"p{p[0]}"
    return p_label.replace("p-", "m").replace("pT1", "T1").replace("p0", "0")


def combine_period(
    period: Tuple[Union[int, str], int],
    window: Tuple[int, int],
    df: pd.DataFrame,
) -> Dict[str, float]:
    """Combines unbiasedness results for a specific period.

    Args:
        period (Tuple[Union[int, str], int]]): Start and end of the period.
        window (Tuple[int, int]]): Start and end of the window.
        df (pd.DataFrame): Unbiasedness regression results for the window.

    Returns:
        Dict[str, float]: Combined unbiasedness results for the period.

    Unbiasedness regression results are dataframes with columns "bench", "d_r2",
    "d_r2_excess" and index corresponding to an event time t.
    """

    p_label = period_prefix(period)

    t1 = window[0]
    t2 = window[1]

    if type(period[0]) == int:
        start = period[0]
    elif period[0] == "T1":
        start = int(df.index[0])
    else:
        raise ValueError(f"Invalid period start: {period[0]}")
    end = period[1]

    if (start < t1) or (end > t2):
        return {
            f"{p_label}_Delta_R2": np.nan,
            f"{p_label}_Delta_Excess_R2": np.nan,
            f"{p_label}_Benchmark": np.nan,
        }

    bench = df.loc[start:end, "bench"].sum()

    out = {
        f"{p_label}_Delta_R2": (d_r2 := df.loc[start:end, "d_r2"].sum()),
        f"{p_label}_Delta_Excess_R2": (d_r2 - bench) / bench,
        f"{p_label}_Benchmark": bench,
    }

    if start != end:
        return out

    return {
        f"{p_label}_R2": float(df.loc[start, "r2"]),
        f"{p_label}_Beta": float(df.loc[start, "beta"]),
        f"{p_label}_SE": float(df.loc[start, "se"]) if "se" in df else np.nan,
        f"{p_label}_VR": float(df.loc[start, "vr"]) if "vr" in df else np.nan,
    } | out


def combine_window(
    window: Tuple[int, int], reg: pd.DataFrame, periods: Iterable[Tuple[int, int]]
) -> Dict[str, float]:
    """
    Combine unbiasedness results for a given window and set of periods.

    Args:
        window (Tuple[int, int]): The start and end times of the window.
        reg (pd.DataFrame): Unbiasedness results for the window.
        periods (Iterable[Tuple[int, int]]): The start and end times of the periods.

    Returns:
        Dict[str, float]: Combined unbiasedness results for the window and periods.

    The `reg` dataframe should have columns "r2", "beta", and index corresponding
    to an event time t. "se", "vr" are optional columns. The function computes the
    unbiasedness results for each period and combines them into a dictionary with keys
    corresponding to the period and window.
    """
    df = reg.copy()
    df["d_r2"] = df["r2"].diff()
    df["d_r2"].iat[0] = df["r2"].iat[0]
    df["bench"] = 1 / len(df)
    df["cumbench"] = df["bench"].cumsum()

    return dict(ChainMap(*[combine_period(p, window, df) for p in periods]))



from typing import List, Union, Callable, Dict, Tuple, Iterable
import datetime as dt
from datetime import date, datetime
from multiprocessing import Pool, cpu_count
import logging

import operator

import pandas as pd
import numpy as np

from .unbiasedness import estim_unbiasedness_regs, estim_unbiasedness_windows
from .windows import get_return_windows, resample_windows


def draw_dates(
    dates: pd.DatetimeIndex,
    N: int,
    dates_pool: pd.DatetimeIndex,
) -> List[pd.DatetimeIndex]:
    """Random draw of matched dates from a given range.

    For each date in dates, draw N dates from the range start_date-end_date,
    with replacement, and matching on day of week and quarter.


    Args:
        dates (pd.DatetimeIndex): Initial dates to match
        N (int): Number of random draws
        dates_pool (pd.DatetimeIndex): Dates to draw from

    Returns:
        List[pd.DatetimeIndex]: List of N random draws of matched dates.
    """

    dr_df = pd.DataFrame(index=dates_pool)
    dr_df["day_of_week"] = dr_df.index.day_of_week
    dr_df["quarter"] = dr_df.index.quarter
    dr_df["year"] = dr_df.index.year

    # For each date, sample from dates in the same quarter and the same day of the
    # week.
    draws = []
    for dt in dates:
        draw = (
            dr_df[
                (dr_df.day_of_week == dt.day_of_week)
                & (dr_df.quarter == dt.quarter)
                & (dr_df.year == dt.year)
            ][[]]
            .sample(n=N, replace=True)
            .reset_index()
        )
        draws.append(draw.T)

    return [
        pd.DataFrame(x, columns=["date"]).set_index("date").index
        for x in np.hsplit(pd.concat(draws).values, N)
    ]


def remove_bad_pvals(df: pd.DataFrame) -> pd.DataFrame:
    """Remove values that are not for Excess Delta R2

    Args:
        df (pd.DataFrame): DataFrame with p-values

    Returns:
        pd.DataFrame: DataFrame with p-values only for Excess Delta R2
    """
    df = df.copy()

    for c in df:
        if not c.endswith("Delta_Excess_R2"):
            df[c] = np.nan

    return df


def shuffles_pvals(res: pd.DataFrame, bs: np.ndarray, n_bs: int = 1000):
    res[[c for c in res.columns if c.endswith("Excess_R2")]]
    bs_pvals = pd.DataFrame(
        np.sum(res.values > bs, axis=0) / n_bs,
        index=res.index,
        columns=res.columns,
    )
    bs_pvals[res.isnull()] = np.nan
    return remove_bad_pvals(bs_pvals)


def bootstrap_pvals(
    res: pd.DataFrame,
    bs: np.ndarray,
    oper: Callable = operator.ge,
    bench: float = 0.0,
    n_bs: int = 1000,
):
    bs_pvals = pd.DataFrame(
        np.sum(oper(bs, bench), axis=0) / n_bs,
        index=res.index,
        columns=res.columns,
    )
    bs_pvals[res.isnull()] = np.nan
    return remove_bad_pvals(bs_pvals)


def _swur_wrapper(args):
    np.random.seed()
    returns, dates, windows, periods, n_bs = args
    logging.info(f"One process for {n_bs} samples running.")
    return [
        estim_unbiasedness_regs(
            returns, dates, windows, periods, use_se=False, shuffle=True
        )
        for _ in range(n_bs)
    ]


def shuffle_window_unbiasedness_reg(
    res: pd.DataFrame,
    returns: pd.Series,
    dates: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    periods: Iterable[Tuple[int, int]],
    n_bs: int = 1000,
    multiprocessing: Union[bool, int] = False,
):
    if multiprocessing:
        n_proc = multiprocessing if (type(multiprocessing) is int) else cpu_count()
        n_proc = min(n_proc, n_bs)
        tasks = [
            (returns, dates, windows, periods, n_bs // n_proc)
            for _ in range(n_proc - 1)
        ]
        tasks += [
            (returns, dates, windows, periods, n_bs - n_bs // n_proc * (n_proc - 1))
        ]
        logging.info(
            f"Launching {n_proc} processes for {n_bs} samples for window shuffles."
        )

        with Pool(processes=n_proc) as pool:
            x = pool.map(_swur_wrapper, tasks)
        shuffle = []
        for s in x:
            shuffle += s
    else:
        shuffle = [
            estim_unbiasedness_regs(
                returns, dates, windows, periods, use_se=False, shuffle=True
            )
            for _ in range(n_bs)
        ]

    return shuffles_pvals(
        res,
        np.array(shuffle),
        n_bs=n_bs,
    )


def _sdur_wrapper(args):
    np.random.seed()
    returns, dates, windows, periods, returns.index, n_bs = args
    logging.info(f"One process for {n_bs} samples running.")
    return [
        estim_unbiasedness_regs(returns, d, windows, periods, use_se=False)
        for d in draw_dates(dates, n_bs, returns.index)
    ]


def shuffle_dates_unbiasedness_reg(
    res: pd.DataFrame,
    returns: pd.Series,
    dates: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    periods: Iterable[Tuple[int, int]],
    n_bs: int = 1000,
    multiprocessing: Union[bool, int] = False,
):
    if multiprocessing:
        n_proc = multiprocessing if (type(multiprocessing) is int) else cpu_count()
        n_proc = min(n_proc, n_bs)
        tasks = [
            (returns, dates, windows, periods, returns.index, n_bs // n_proc)
            for _ in range(n_proc - 1)
        ]
        remainder = n_bs - n_bs // n_proc * (n_proc - 1)
        tasks += [(returns, dates, windows, periods, returns.index, remainder)]
        logging.info(
            f"Launching {n_proc} processes for {n_bs} samples for dates shuffles."
        )

        with Pool(processes=n_proc) as pool:
            x = pool.map(_sdur_wrapper, tasks)
        shuffle = []
        for s in x:
            shuffle += s
    else:
        shuffle = [
            estim_unbiasedness_regs(returns, d, windows, periods, use_se=False)
            for d in draw_dates(dates, n_bs, returns.index)
        ]

    return shuffles_pvals(
        res,
        np.array(shuffle),
        n_bs=n_bs,
    )


def _bur_wrapper(args):
    np.random.seed()
    win_dfs, periods, n_bs = args
    logging.info(f"One process for {n_bs} samples running.")
    return [
        estim_unbiasedness_windows(resample_windows(win_dfs), periods, use_se=False)
        for _ in range(n_bs)
    ]


def bootstrap_unbiasedness_regs(
    res: pd.DataFrame,
    returns: pd.Series,
    dates: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    periods: Iterable[Tuple[int, int]],
    n_bs: int = 1000,
    multiprocessing: Union[bool, int] = False,
) -> np.array:
    win_dfs = get_return_windows(returns, dates, windows)

    # Each iteration, resample with replacement
    if multiprocessing:
        bootstrap = compute_bs_pvals_mp(multiprocessing, n_bs, win_dfs, periods)
    else:
        bootstrap = [
            estim_unbiasedness_windows(resample_windows(win_dfs), periods, use_se=False)
            for _ in range(n_bs)
        ]

    return bootstrap_pvals(
        res,
        np.array(bootstrap),
        n_bs=n_bs,
    )


def compute_bs_pvals_mp(
    multiprocessing: Union[bool, int],
    n_bs: int,
    win_dfs: Dict[str, pd.DataFrame],
    periods: Iterable[Tuple[int, int]],
) -> List[pd.DataFrame]:
    n_proc = multiprocessing if (type(multiprocessing) is int) else cpu_count()
    n_proc = min(n_proc, n_bs)
    tasks = [(win_dfs, periods, n_bs // n_proc) for _ in range(n_proc - 1)]
    remainder = n_bs - n_bs // n_proc * (n_proc - 1)
    tasks += [(win_dfs, periods, remainder)]
    logging.info(f"Launching {n_proc} processes for {n_bs} samples for bootstrap.")

    with Pool(processes=n_proc) as pool:
        x = pool.map(_bur_wrapper, tasks)
    result = []
    for s in x:
        result += s
    return result


def estim_unbiasedness_regs_bs(
    returns: pd.Series,
    dates: pd.DatetimeIndex,
    windows: Iterable[Tuple[int, int]],
    periods: Iterable[Tuple[int, int]],
    shuffle_window=False,
    shuffle_dates=False,
    bootstrap=True,
    n_bs=1000,
    multiprocessing: Union[bool, int] = False,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    if n_bs <= 0:
        raise ValueError(
            f"The number of boostrap samples (n_bs) should be positive: {n_bs}"
        )

    returns = returns.dropna()

    # Keep only dates that are in the returns
    dates = returns.index.intersection(dates)

    # Estimate unbiasedness regressions
    res = estim_unbiasedness_regs(returns, dates, windows, periods)

    pvals = []
    # Shuffles
    # First shuffle: keep the same dates/windows, but shuffle dates within
    # each window.
    if shuffle_window:
        shuffle_window_vals = shuffle_window_unbiasedness_reg(
            res=res,
            returns=returns,
            dates=dates,
            windows=windows,
            periods=periods,
            n_bs=n_bs,
            multiprocessing=multiprocessing,
        )
        pvals.append(shuffle_window_vals)

    # Second shuffle: draw new dates, matching on year/quarter/day of week
    if shuffle_dates:
        shuffle_dates_vals = shuffle_dates_unbiasedness_reg(
            res=res,
            returns=returns,
            dates=dates,
            windows=windows,
            periods=periods,
            n_bs=n_bs,
            multiprocessing=multiprocessing,
        )
        pvals.append(shuffle_dates_vals)

    # # Bootstrap
    if bootstrap:
        bootstrap_vals = bootstrap_unbiasedness_regs(
            res=res,
            returns=returns,
            dates=dates,
            windows=windows,
            periods=periods,
            n_bs=n_bs,
            multiprocessing=multiprocessing,
        )
        pvals.append(bootstrap_vals)

    return tuple([res] + pvals)
